import { currMouseAction, backToEdit } from "../menutools.js"
import { MouseAction, ElementType } from "./Enums.js";
import { Node, fillValue } from "./Node.js";
import { colorMouseOver, fileManager } from "../simulator.js"
import { checkCollision, colorCollision } from "../simulator.js";



/**
 * Generate input for the circuit
 * @classdesc Generate input for the circuit
 * @memberof module:LogicInput
 */
export class LogicInput {

    /**
     * Contructor
     */
    constructor() {

        this.value = false;
        this.posX = mouseX;
        this.posY = mouseY;
        this.diameter = 25;
        this.collided = false;
        this.isSpawned = false;
        this.isMoving = false;
        this.offsetMouseX = 0;
        this.offsetMouseY = 0;
        this.output = new Node(this.posX + 30, this.posY, true, this.value);
        this.nodeStartID = this.output.id;
        this.isSaved = false;
    }
    

    /**
     * Create a LogicInput from json
     * @param {*} json Json File
     * @returns {LogicInput} LogicInput generate from json
     */
    static from(json){
        return Object.assign(this, json);
    }

    /**
     * Destroy this LogicInput
     */
    destroy() {
        this.output.destroy();
        delete this.output;
    }

    /**
     * Draw this LogicInput
     */
    draw() {
        if (!this.isSpawned) {
            this.posX = mouseX;
            this.posY = mouseY;
        }else if(!this.isSaved)
        {
            fileManager.saveState();
            this.isSaved = true;
        }

        fillValue(this.value);

        if (this.isMoving) {
            this.posX = mouseX + this.offsetMouseX;
            this.posY = mouseY + this.offsetMouseY;
        }

        if(this.isMoving || !this.isSpawned){
            //If is moving or is being spawed ceck collision
            this.collided = checkCollision(this);
        }
        this.drawBoundaries();


        this.output.updatePosition(this.posX + 30, this.posY);
        this.output.setValue(this.value);
        this.output.draw();

        this.printInfo();

        textSize(18);

        if (this.value) {
            textStyle(BOLD);
            text('1', this.posX - this.diameter / 4, this.posY + this.diameter / 4);
        }
        else {
            textStyle(NORMAL);
            fill(255);
            text('0', this.posX - this.diameter / 4, this.posY + this.diameter / 4);
        }
    }

    /**
     * Draw element doundaries
     */
    drawBoundaries(){
        if(!this.collided && this.isMouseOver()){
            stroke(colorMouseOver[0], colorMouseOver[1], colorMouseOver[2]);
        }else if(this.collided){
            stroke(colorCollision[0], colorCollision[1], colorCollision[2]);
        }else{
            stroke(0);
        }
        strokeWeight(4);
        line(this.posX, this.posY, this.posX + 30, this.posY);
        circle(this.posX, this.posY, this.diameter);
    }

    /**
     * Refresh this LogicInput node ID
     */
    refreshNodes()
    {
        let currentID = this.nodeStartID;
        this.output.setID(currentID);
    }


    /**
     * Print this LogicInput info
     */
    printInfo() {
        noStroke();
        fill(0);
        textSize(12);
        textStyle(NORMAL);
        text('LOG. INPUT', this.posX - 20, this.posY + 25);
    }

    /**
     * Checking if mouse if over the input
     * @returns {Boolean} Boolean Value
     */
    isMouseOver() {
        if (dist(mouseX, mouseY, this.posX, this.posY) < this.diameter / 2)
            return true;
        return false;
    }

    /**
     * Called when mouse is pressed
     * If the element is not spawned, it will be spawned.
     * Then if mouse is over OR the current action is MOVE then move it
     */
    mousePressed() {
        if (!this.isSpawned  && !checkCollision(this)) {
            this.posX = mouseX;
            this.posY = mouseY;
            this.isSpawned = true;
            backToEdit();
            return;
        }

        if (this.isMouseOver() || currMouseAction == MouseAction.MOVE) {
            this.isMoving = true;
            this.offsetMouseX = this.posX - mouseX;
            this.offsetMouseY = this.posY - mouseY;
            this.posXOld = this.posX;
            this.posYOld = this.posY;
        }
    }

    /**
     * Called when mouse is released
     * If the element was moving, release it. 
     */
    mouseReleased() {  
        if(this.isMoving && (this.collided=checkCollision(this))){
            this.posX = this.posXOld;
            this.posY = this.posYOld
        } 
        this.isMoving = false;         
    }

    /**
     * Called when mouse is double clicked
     * If mouse is over this instance
     * Invert input value 
     */
    doubleClicked() {
        if (this.isMouseOver())
            this.value ^= true;
    }

    /**
     * Called when mouse is clicked
     * @returns {Boolean} Boolean
     */
    mouseClicked() {
        if (this.isMouseOver() || this.output.isMouseOver()) {
            this.output.mouseClicked();
            return true;
        }
        return false;
    }

    /**
     * Invert this LogicInput value
     */
    toggle() {
        this.value ^= true;
    }

    /**
     * Check if this element has collision with Coordinate
     * @param {Int[]} coordinate Coordinate to check collision
     * @returns {Boolean} Boolean value
     */
    checkCollision(coordinate){
        if((coordinate[0] + coordinate[2] < this.posX - (this.diameter/2) || coordinate[0] > this.posX + (this.diameter/2) || coordinate[1] + coordinate[3] < this.posY - (this.diameter/2) || coordinate[1] > this.posY + (this.diameter/2) )){
            return false;
        }
        return true;
    }

    /**
     * Return element position and dimension
     * @returns {Array} Array with [ posx , posy , width , height]
     */
    getCoordinate(){
        let coordinate = [this.posX - (this.diameter/2),this.posY - (this.diameter/2),this.diameter,this.diameter];
        return coordinate
    }

}

